from django.contrib import messages
from invitations.adapters import get_invitations_adapter
from invitations.app_settings import app_settings as invitations_settings
from invitations.signals import invite_accepted
from xlibrary.viewsets import ModelViewset as xViewSet
from rest_framework import mixins, status, viewsets
from rest_framework.decorators import (action, api_view,
                                       permission_classes)
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.response import Response
from .models import GuestInvitation as InvitationModel, GroupOptions, User
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
# from .serializers import InvitationReadSerializer,InvitationBulkWriteSerializer,InvitationWriteSerializer,UserSerializer, GroupSerializer, PermissionSerializer,ContenttypeSerializer,GroupOptionsSerializer,CUserSerializer, CGroupSerializer,CGroupOptionsSerializer
from .serializers import InvitationReadSerializer, InvitationBulkWriteSerializer, InvitationWriteSerializer, UserSerializer, GroupSerializer, PermissionSerializer, ContenttypeSerializer, GroupOptionsSerializer, CUserSerializer, CGroupSerializer
from rest_auth.views import LoginView as RestAuthLoginView
from django.conf import settings
from .utils import get_author_model
from django.db.models.functions import Concat
AuthorModel = get_author_model()


def getVisiblePermissions():
    ct = ContentType.objects.all()
    myNameList = []
    for c in ct:
        class_name = c.model_class()
        if hasattr(class_name, "xtrmMeta"):
            myNameList.append((class_name._meta.app_label +
                               class_name.__name__).lower())
    return myNameList


def jwt_response_payload_handler(token, user=None, request=None):
    myNameList = getVisiblePermissions()
    perms = user.groups.first()
    rtn = []
    if perms:
        for perm in perms.permissions.all().annotate(num_authors=Concat('content_type__app_label', 'content_type__model')).filter(num_authors__in=myNameList):
            pm = {
                "id": perm.id,
                "name": perm.name,
                "codename": perm.codename,
                "app_label": perm.content_type.app_label,
                "model": perm.content_type.model
            }
            rtn.append(pm)

    return {
        'token': token,
        'user': UserSerializer(user, context={'request': request}).data,
        'permissions': rtn
    }


class LoginView(RestAuthLoginView):
    """
    Overriden LoginView

    get_response serializes the response according to the default serializer (as you already know)
    which in turn, uses the default UserDetailsSeriazlizer (which we don't want)

    What we can do is use our custom jwt_response_payload_handler as our serializer and return in directly
    in the response
    """

    def get_response(self):
        serializers_class = self.get_response_serializer()

        if getattr(settings, 'REST_USE_JWT', False):

            # Use jwt_response_payload_handler since we already have the token, user (optional) and the request (optional)
            data = jwt_response_payload_handler(
                self.token, self.user, self.request)

            return Response(data, status=status.HTTP_200_OK)
        else:
            serializer = serializers_class(instance=self.token, context={
                                           'resquest': self.request})

            return Response(serializer.data, status=status.HTTP_200_OK)


class UserViewSet(xViewSet):
    #http_method_names = [m for m in viewsets.ModelViewSet.#http_method_names if m not in [m for m in viewsets.ModelViewSet.#http_method_names if m not in ['get']]]
    queryset = User.objects.all()
    serializer_class = UserSerializer
    reporttitle = 'Users'
#orientation='landscape'
    options = {'canAdd': 1, 'canPrint': 1}
    filters = [
        {'title': 'User Name', 'name': 'filter{username}', 'url': 'v1/rights/users/'},
        {'title': 'Email', 'name': 'filter{email}'},
        {"title": "Group", "name": "filter{groups.name}", "url": "v1/rights/groups/"}
    ]
    columns = [
        {'title': 'Id', 'name': 'id', 'type': 'numeric',
            'searchable': False, 'sortable': False, 'visible': False},
        {'title': 'User Name', 'name': 'username', 'width': '20%'},
        {'title': 'First Name', 'name': 'first_name', 'width': '20%'},
        {'title': 'Last Name', 'name': 'last_name', 'width': '20%'},
        {'title': 'Email', 'name': 'email'}
    ]


class GroupViewSet(xViewSet):
    #http_method_names = [m for m in viewsets.ModelViewSet.#http_method_names if m not in [m for m in viewsets.ModelViewSet.#http_method_names if m not in ['get']]]
    queryset = Group.objects.all()
    serializer_class = GroupSerializer
    reporttitle = 'Groups'
    #orientation='landscape'
    options = {'canAdd': 1, 'canPrint': 1}
    filters = [
        {'title': 'name', 'name': 'filter{name}', 'url': 'v1/rights/groups/'}
    ]
    columns = [
        {'title': 'Id', 'name': 'id', 'alignment': 'right', 'type': 'numeric',
            'searchable': False, 'sortable': False, 'visible': False},
        {'title': 'Name', 'name': 'name'}
    ]


class PermissionViewSet(xViewSet):
    #http_method_names = [m for m in viewsets.ModelViewSet.#http_method_names if m not in [m for m in viewsets.ModelViewSet.#http_method_names if m not in ['get']]]
    queryset = Permission.objects.all()
    serializer_class = PermissionSerializer
    reporttitle = 'Permissions'
    #orientation='landscape'
    options = {'canAdd': 1, 'canPrint': 1}
    filters = [
        {'title': 'Name', 'name': 'filter{name}', 'url': 'v1/rights/permissions/'}
    ]
    columns = [
        {'title': 'Id', 'name': 'id', 'alignment': 'right', 'type': 'numeric',
            'searchable': False, 'sortable': False, 'visible': False},
        {'title': 'Name', 'name': 'name'},
        {'title': 'Code Name', 'name': 'codename'}
    ]

    def get_queryset(self):
        myNameList = getVisiblePermissions()
        return Permission.objects.all().filter(content_type__model__in=myNameList)


class ContenttypeViewSet(xViewSet):
    #http_method_names = [m for m in viewsets.ModelViewSet.#http_method_names if m not in [m for m in viewsets.ModelViewSet.#http_method_names if m not in ['get']]]
    queryset = ContentType.objects.all()
    serializer_class = ContenttypeSerializer
    reporttitle = 'Content Types'
    #orientation='landscape'
    options = {'canAdd': 1, 'canPrint': 1}
    filters = [
        {'title': 'App Label', 'name': 'filter{app_label}',
            'url': 'v1/rights/contenttypes/'}
    ]
    columns = [
        {'title': 'Id', 'name': 'id', 'alignment': 'right', 'type': 'numeric',
            'searchable': False, 'sortable': False, 'visible': False},
        {'title': 'App Label', 'name': 'app_label'},
        {'title': 'Model', 'name': 'model'}
    ]


class GroupOptionsViewSet(xViewSet):
    #http_method_names = [m for m in viewsets.ModelViewSet.#http_method_names if m not in [m for m in viewsets.ModelViewSet.#http_method_names if m not in ['get']]]
    queryset = GroupOptions.objects.all()
    serializer_class = GroupOptionsSerializer


class CUserViewSet(viewsets.ModelViewSet):
    #http_method_names = [m for m in viewsets.ModelViewSet.#http_method_names if m not in ['get']]
    queryset = User.objects.all()
    serializer_class = CUserSerializer

    # def get_permissions(self):
    #     permission_classes = [IsAdminUser]
    #     return [permission() for permission in permission_classes]


class CGroupViewSet(viewsets.ModelViewSet):
    #http_method_names = ['post', 'put', 'patch']
    queryset = Group.objects.all()
    serializer_class = CGroupSerializer


# class CGroupOptionsViewSet(viewsets.ModelViewSet):
#     queryset = GroupOptions.objects.all()
#     serializer_class = CGroupOptionsSerializer

class InvitationViewSet(viewsets.ModelViewSet):
    queryset = InvitationModel.objects.all()
    permission_classes = [IsAuthenticated]

    def get_serializer_class(self):
        if self.action in ['list', 'retrieve']:
            return InvitationReadSerializer
        elif self.action == 'send_multiple':
            return InvitationBulkWriteSerializer
        return InvitationWriteSerializer

    def _prepare_and_send(self, invitation, request):
        invitation.inviter = request.user
        invitation.save()
        invitation.send_invitation(request)

    # @detail_route(
    #     methods=['post'], permission_classes=[IsAuthenticated],
    #     url_path=SEND_URL
    # )
    # def send(self, request, pk=None):
    #     invitation = self.get_object()
    #     self._prepare_and_send(invitation, request)
    #     content = {'detail': 'Invite sent'}
    #     return Response(content, status=status.HTTP_200_OK)

    # @list_route(
    #     methods=['post'], permission_classes=[IsAuthenticated],
    #     url_path=CREATE_AND_SEND_URL
    # )
    @action(
        methods=['post'], detail=False
    )
    def send(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        email = serializer.data['email']
        authorentity = AuthorModel.objects.get(
            pk=serializer.data['authorentity'])
        invitation = InvitationModel.create(
            email=email, authorentity=authorentity, inviter=request.user)
        self._prepare_and_send(invitation, request)
        content = {'detail': 'Invite sent'}
        return Response(content, status=status.HTTP_200_OK)

    # @list_route(
    #     methods=['post'], permission_classes=[IsAuthenticated],
    #     url_path=SEND_MULTIPLE_URL
    # )
    @action(
        methods=['post'], detail=False
    )
    def sendmultiple(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        inviter = request.user
        authorentity = AuthorModel.objects.get(
            pk=serializer.data['authorentity'])
        for email in serializer.data['email']:
            invitation = InvitationModel.create(
                email=email, authorentity=authorentity, inviter=inviter)
            self._prepare_and_send(invitation, request)
        content = {'detail': 'Invite(s) sent'}
        return Response(content, status=status.HTTP_200_OK)


INVALID_INVITE = 'invalid-invite'
ACCEPTED_INVITE = 'accepted-invite'
EXPIRED_INITE = 'expired-invite'
SUCCESS_INVITE = 'success'


def verify_invitation_object(invitation):
    if not invitation:
        return {INVALID_INVITE: 'Invalid/Expired Invitation.'}

    if invitation.accepted:
        return {ACCEPTED_INVITE: 'This invitation has already been accepted.'}

    if invitation.key_expired():
        return {EXPIRED_INITE: 'Invitation has expired.'}

    return {SUCCESS_INVITE: 'This is a valid invitation'}


@api_view(['GET'])
@permission_classes((AllowAny,))
def accept_invitation(request, key):
    def get_object():
        try:
            return InvitationModel.objects.get(key=key.lower())
        except InvitationModel.DoesNotExist:
            return None

    invitation = get_object()

    verified_data = verify_invitation_object(invitation)

    return Response(
        verified_data,
        status=status.HTTP_200_OK
    )


@api_view(['POST'])
@permission_classes((AllowAny,))
def signup(request, key):
    def get_object():
        try:
            return InvitationModel.objects.get(key=key.lower())
        except InvitationModel.DoesNotExist:
            return None

    invitation = get_object()
    INVITE_STATUS = verify_invitation_object(invitation)
    if SUCCESS_INVITE in INVITE_STATUS:
        if invitation.authorentity.guestlogin is None:
            request.data['group_id'] = "-1"
            serialized = CUserSerializer(data=request.data)
            serialized.is_valid(raise_exception=True)
            user = serialized.save()
            invitation.authorentity.guestlogin = user
            invitation.authorentity.save()
            invitation.accepted = True
            invitation.save()
            invite_accepted.send(sender=None, email=invitation.email)
            get_invitations_adapter().add_message(
                request,
                messages.SUCCESS,
                'invitations/messages/invite_accepted.txt',
                {
                    'email': invitation.email,
                    'author': invitation.authorentity.pk
                }
            )
        else:
            return Response(
                {INVALID_INVITE: AuthorModel._meta.verbose_name +
                    ' has already been assigned a Guest Account.'},
                status=status.HTTP_200_OK
            )
    elif INVALID_INVITE in INVITE_STATUS:
        get_invitations_adapter().add_message(
            request,
            messages.ERROR,
            'invitations/messages/invite_invalid.txt'
        )
    elif ACCEPTED_INVITE in INVITE_STATUS:
        get_invitations_adapter().add_message(
            request,
            messages.ERROR,
            'invitations/messages/invite_already_accepted.txt',
            {
                'email': invitation.email
            }
        )
    elif EXPIRED_INITE in INVITE_STATUS:
        get_invitations_adapter().add_message(
            request,
            messages.ERROR,
            'invitations/messages/invite_expired.txt',
            {
                'email': invitation.email
            }
        )
    return Response(
        INVITE_STATUS,
        status=status.HTTP_200_OK
    )



from .models import GuestInvitation as InvitationModel, GroupOptions, User
# from django.contrib.auth import get_user_model
from django.utils.translation import ugettext_lazy as _
from invitations.adapters import get_invitations_adapter
from invitations.exceptions import (AlreadyAccepted, AlreadyInvited,
                                    UserRegisteredEmail)
from rest_framework import serializers
from xtrm_drest.serializers import (
    DynamicModelSerializer,
    DynamicRelationField
)
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from .utils import get_author_model
errors = {
    "already_invited": _("This e-mail address has already been"
                         " invited."),
    "already_accepted": _("This e-mail address has already"
                          " accepted an invite."),
    "email_in_use": _("An active user is using this e-mail address"),
}

AuthorModel = get_author_model()


class EmailListField(serializers.ListField):
    child = serializers.EmailField()


class InvitationWriteSerializer(serializers.ModelSerializer):
    class Meta:
        model = InvitationModel
        fields = ('email', 'authorentity',)

    def _validate_invitation(self, email):
        if InvitationModel.objects.all_valid().filter(
                email__iexact=email, accepted=False):
            raise AlreadyInvited
        elif InvitationModel.objects.filter(
                email__iexact=email, accepted=True):
            raise AlreadyAccepted
        elif User.objects.filter(email__iexact=email):
            raise UserRegisteredEmail
        else:
            return True

    def validate_authorentity(self, authorentity):
        if isinstance(authorentity, AuthorModel):
            author = authorentity
        else:
            try:
                author = AuthorModel.objects.get(pk=authorentity)
            except AuthorModel.DoesNotExist:
                author = None
        if author is None:
            raise serializers.ValidationError(
                ['Please specify a valid ' + AuthorModel._meta.verbose_name])
        elif not author.guestlogin is None:
            raise serializers.ValidationError(
                [AuthorModel._meta.verbose_name + ' has already been assigned a Guest Account'])

    def validate_email(self, email):
        email = get_invitations_adapter().clean_email(email)

        try:
            self._validate_invitation(email)
        except(AlreadyInvited):
            raise serializers.ValidationError(errors["already_invited"])
        except(AlreadyAccepted):
            raise serializers.ValidationError(errors["already_accepted"])
        except(UserRegisteredEmail):
            raise serializers.ValidationError(errors["email_in_use"])
        return email

    def create(self, validate_data):
        return InvitationModel.create(**validate_data)


class InvitationReadSerializer(serializers.ModelSerializer):
    class Meta:
        model = InvitationModel
        fields = '__all__'


class InvitationBulkWriteSerializer(InvitationWriteSerializer):

    email = EmailListField()

    class Meta(InvitationWriteSerializer.Meta):
        model = InvitationModel
        fields = InvitationWriteSerializer.Meta.fields

    def validate_email(self, email_list):
        if len(email_list) == 0:
            raise serializers.ValidationError(
                _('You must add one or more email addresses')
            )
        for email in email_list:
            email = get_invitations_adapter().clean_email(email)
            try:
                self._validate_invitation(email)
            except(AlreadyInvited):
                raise serializers.ValidationError(errors["already_invited"])
            except(AlreadyAccepted):
                raise serializers.ValidationError(errors["already_accepted"])
            except(UserRegisteredEmail):
                raise serializers.ValidationError(errors["email_in_use"])
            return email_list


class ContenttypeSerializer(DynamicModelSerializer):
    class Meta:
        model = ContentType
        fields = '__all__'


class GroupSerializer(DynamicModelSerializer):
    permissions = DynamicRelationField("PermissionSerializer", many=True)

    class Meta:
        model = Group
        fields = '__all__'


class UserSerializer(DynamicModelSerializer):
    groups = DynamicRelationField(GroupSerializer, many=True)
    class Meta:

        model = User
        exclude = ('user_permissions', 'password',)


class PermissionSerializer(DynamicModelSerializer):
    content_type = DynamicRelationField("ContenttypeSerializer")

    class Meta:
        model = Permission
        fields = '__all__'


class GroupOptionsSerializer(DynamicModelSerializer):
    groups = DynamicRelationField(GroupSerializer, writable=True)

    class Meta:
        model = GroupOptions
        fields = '__all__'


class CUserSerializer(serializers.ModelSerializer):
    group_id = serializers.CharField(write_only=True)
    # author = serializers.IntegerField(write_only=True,required=False)

    class Meta:
        exclude = ('user_permissions',)
        read_only_fields = ('last_login', 'is_superuser',
                            'is_staff', 'date_joined', 'groups')
        model = User
        extra_kwargs = {'password': {'write_only': True}}

    def validate(self, data):
        if self.instance and self.instance.is_staff == False:
            raise serializers.ValidationError(
                {'username': ['You can not modify Guests']})
        # if data['author']:
            # if 'email' not in data or not data['email']:
            #     raise serializers.ValidationError(
            #         {'email': ['Email ID is mandatory for Guests']})
            # try:
            #     guestObj = InvitationModel.objects.get(email=data['email'])
            # except InvitationModel.DoesNotExist:
            #     guestObj=None
            # if not guestObj:
            #     raise serializers.ValidationError(
            #         {'email': ['No invitation has been sent to this Email ID']})
            # elif not guestObj.accepted:
            #     raise serializers.ValidationError(
            #         {'email': ['You need to accept invitation for this Email ID, Before signing up']})
            # try:
            #     authorObj = AuthorModel.objects.get(pk=data['author'])
            # except AuthorModel.DoesNotExist:
            #     authorObj=None
            # if not authorObj:
            #     raise serializers.ValidationError({'author': ['Invalid ' + AuthorModel._meta.verbose_name]})
            # elif not authorObj.guestlogin is None:
            #     raise serializers.ValidationError(
            #         {'author': ['A Guest User has already been assigned to this ' + AuthorModel._meta.verbose_name]})
        # elif 'group_id' in data and data['group_id']=="2":
        if 'group_id' in data and data['group_id'] == "2":
            raise serializers.ValidationError(
                {'username': ['You can not create a Guest User, You need to invite him/her']})
        return data

    def create(self, validated_data):
        group_id = validated_data.pop('group_id', None)
        # author_id = validated_data.pop('author', None)
        if group_id == "-1":
            group_id = "2"
        user = User.objects.create(**validated_data)
        user.is_active = True
        user.is_superuser = False
        if group_id == "2":
            user.is_staff = False
        else:
            user.is_staff = True
        user.set_password(validated_data['password'])
        user.groups.add(group_id)
        user.save()
        # if author_id:
        #     authorObj=AuthorModel.objects.get(pk=author_id)
        #     if authorObj:
        #         authorObj.guestlogin=user
        #         authorObj.save()
        return user

    def update(self, instance, validated_data):
        group_id = validated_data.pop('group_id', None)
        user = super(CUserSerializer, self).update(instance, validated_data)
        user.is_superuser = False
        if group_id == 2:
            user.is_staff = False
        else:
            user.is_staff = True
        user.groups.clear()
        user.groups.add(group_id)
        user.save()
        return user


class CGroupSerializer(serializers.ModelSerializer):
    privilege = serializers.CharField(
        source='groups.privilege', read_only=True)
    option = serializers.CharField(write_only=True)

    def associate_permission(self, permission, group):
        ct_obj = set(p.content_type.model_class() for p in permission if hasattr(p.content_type.model_class(
        ), "xtrmMeta") and hasattr(p.content_type.model_class().xtrmMeta, "associate_model"))
        if ct_obj:
            asList = []
            for mclass in ct_obj:
                pList = tuple(
                    p.codename for p in permission if p.content_type.model_class() == mclass)
                for p in pList:
                    x = p.split('_')
                    for m in mclass.xtrmMeta.associate_model:
                        asList.append(x[0] + '_' + m.lower())
            if asList and len(asList) > 0:
                permission.extend(list(Permission.objects.filter(
                    codename__in=asList).values_list('id', flat=True)))
        group.permissions.set(permission)

    def create(self, validated_data):
        opts = validated_data.get('option')
        if opts:
            opts = validated_data.pop('option')
        perms = validated_data.get('permissions')
        if perms:
            perms = validated_data.pop('permissions')
        gp = Group.objects.create(**validated_data)
        if perms:
            self.associate_permission(perms, gp)
        if opts:
            GroupOptions.objects.create(groups=gp, privilege=opts)
        return gp

    def update(self, instance, validated_data):
        opts = validated_data.get('option')
        if opts:
            opts = validated_data.pop('option')
        perms = validated_data.get('permissions')
        if perms:
            perms = validated_data.pop('permissions')
        oinstance = super(CGroupSerializer, self).update(
            instance, validated_data)
        oinstance.permissions.clear()
        if perms:
            self.associate_permission(perms, oinstance)
        groupOption = GroupOptions.objects.get(groups=oinstance)
        if groupOption and opts:
            groupOption.privilege = opts
            groupOption.save()
        return oinstance

    class Meta:
        model = Group
        name = 'group'
        fields = '__all__'


# class CGroupOptionsSerializer(serializers.ModelSerializer):
#     groups = GroupSerializer(required=False, read_only=False)

#     class Meta:
#         model = GroupOptions
#         fields = '__all__'



# from rest_framework import routers
from .views import InvitationViewSet, accept_invitation, UserViewSet, GroupViewSet, PermissionViewSet, ContenttypeViewSet, GroupOptionsViewSet, CUserViewSet, CGroupViewSet, signup
from django.conf.urls import include, url
from xtrm_drest.routers import DynamicRouter

# router = routers.SimpleRouter()
router = DynamicRouter()
router.register('users', UserViewSet)
router.register('edit/user', CUserViewSet)
router.register('groups', GroupViewSet)
router.register('edit/group', CGroupViewSet)
router.register('permissions', PermissionViewSet)
router.register('contenttypes', ContenttypeViewSet)
router.register('groupoptions', GroupOptionsViewSet)
router.register('invitations', InvitationViewSet)
invitations_patterns = (
    [
        url(
            r'^{0}/{1}/(?P<key>\w+)/?$'.format(
                'invitations', 'verify'
            ),
            accept_invitation,
            name='accept-invite'
        ),
        url(
            r'^{0}/{1}/(?P<key>\w+)/?$'.format(
                'invitations', 'signup'
            ),
            signup,
            name='signup-invite'
        ),
    ],
    'invitations'
)

urlpatterns = router.urls + [
    url(r'^', include(invitations_patterns)),
]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jevan-library", # Replace with your own username
    version="0.0.9",
    include_package_data=True,
    author="Jevan Solutions",
    author_email="jevan.xtrm@gmail.com",
    description="Core library for each project developed by Jevan Solutions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="", # https://github.com/pypa/sampleproject
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)